from flask import Flask, render_template
from time import sleep
import multiprocessing
import traceback
import threading
import logging
import os

from ..libraries.runtime_args import RuntimeArgs, parse_args
from ..libraries.version_manager import is_update_available, get_installed_version, lookup_latest_version
from ..libraries.app_scope import is_local_run

app = Flask(
    __name__
)
log = logging.getLogger('core')

## Import and register BPs
################################

from .blueprints.api import api_bp
from .blueprints.web import web_bp

app.register_blueprint(api_bp)
app.register_blueprint(web_bp)
    
## Define global jinja filters
################################

def is_substring_in_values(results: dict, substring: str) -> bool:
    return any(substring.lower() in str(v).lower() for v in results.values()) if substring else True

app.jinja_env.filters['is_substring_in_values'] = is_substring_in_values

## Define global jinja vars
################################

def set_global_safe(key: str, value):
    """ Safely set global vars without worrying about an exception """
    app_globals = app.jinja_env.globals
    try:
        if callable(value): value = value()
        
        app_globals[key] = value
        log.debug(f'jinja_globals[{key}] = {value}')
    except:
        default = app_globals.get(key)
        log.debug(traceback.format_exc())
        log.info(
            f"Unable to set app global var '{key}'"+
            f"defaulting to '{default}'"
        )
        app_globals[key] = default

set_global_safe('app_version',get_installed_version)
set_global_safe('update_available', is_update_available)
set_global_safe('latest_version',lookup_latest_version)
set_global_safe('runtime_args', vars(parse_args()))
set_global_safe('is_local',is_local_run)

## External hook to kill flask server
################################

exiting = False
@app.route("/shutdown")
def exit_app():
    global exiting
    exiting = True
    log.info('Received external exit request. Terminating flask.')
    return "Done"

@app.teardown_request
def teardown(exception):
    if exiting:
        os._exit(0)

## Generalized error handling
################################
@app.errorhandler(500)
def internal_error(e):
    """
    handle internal errors nicely
    """
    tb = traceback.format_exc()
    return render_template('error.html',
                           error=None,
                           traceback=tb), 500

## Webserver creation functions
################################

def start_webserver_dameon(args: RuntimeArgs) -> multiprocessing.Process:
    proc = threading.Thread(target=start_webserver, args=(args,))
    proc.daemon = True # Kill thread when main thread exits
    proc.start()
    log.info('Flask server initializing as dameon')
    sleep(2)

def start_webserver(args: RuntimeArgs) -> int:
    app.run(
        host='0.0.0.0', 
        port=args.port, 
        debug=args.reloader,
        use_reloader=args.reloader
    )


if __name__ == "__main__":
    start_webserver(True)





import logging 
import requests
import traceback
import pkg_resources
from random import randint

from .app_scope import is_local_run

log = logging.getLogger('VersionManager')

PACKAGE='lanscape'
LOCAL_VERSION = '0.0.0'

latest = None # used to 'remember' pypi version each runtime

def is_update_available(package=PACKAGE) -> bool:
    installed = get_installed_version(package)
    available = lookup_latest_version(package)
    if installed == LOCAL_VERSION: return False #local

    return installed != available

def lookup_latest_version(package=PACKAGE):
    # Fetch the latest version from PyPI
    global latest
    if not latest:
        no_cache = f'?cachebust={randint(0,6969)}'
        url = f"https://pypi.org/pypi/{package}/json{no_cache}"
        try:
            response = requests.get(url)
            response.raise_for_status()  # Raise an exception for HTTP errors
            latest = response.json()['info']['version']
            log.debug(f'Latest pypi version: {latest}')
        except:
            log.debug(traceback.format_exc())
            log.warning('Unable to fetch package version from PyPi')
    return latest

def get_installed_version(package=PACKAGE):
    if not is_local_run():
        try:
            return pkg_resources.get_distribution(package).version
        except:
            log.debug(traceback.format_exc())
            log.warning(f'Cannot find {package} installation')
    return LOCAL_VERSION
    




from time import time
from dataclasses import dataclass, field
from typing import DefaultDict
from collections import defaultdict
from tabulate import tabulate
@dataclass
class JobStats:
    running: DefaultDict[str, int] = field(default_factory=lambda: defaultdict(int))
    finished: DefaultDict[str, int] = field(default_factory=lambda: defaultdict(int))
    timing: DefaultDict[str, float] = field(default_factory=lambda: defaultdict(float))

    def __str__(self):
        data = [
            [name, self.running.get(name, 0), self.finished.get(name, 0), self.timing.get(name, 0.0)]
            for name in set(self.running) | set(self.finished)
        ]

        headers = ["Function", "Running", "Finished", "Avg Time (s)"]
        return tabulate(data, headers=headers, tablefmt="grid")


def job_tracker(func):
    fxn = func.__name__

    def wrapper(*args, **kwargs):
        # Access the class instance and initialize job tracking stats
        class_instance = args[0]
        job_stats = init_job_tracker(class_instance)

        # Increment running counter and track execution time
        job_stats.running[fxn] += 1
        start = time()
        
        result = func(*args, **kwargs)  # Execute the wrapped function
        
        # Update statistics after function execution
        elapsed = time() - start
        job_stats.running[fxn] -= 1
        job_stats.finished[fxn] += 1

        # Calculate the new average timing for the function
        job_stats.timing[fxn] = round(
            ((job_stats.finished[fxn] - 1) * job_stats.timing[fxn] + elapsed) / job_stats.finished[fxn], 4
        )

        # Clean up if no more running instances of this function
        if job_stats.running[fxn] == 0:
            job_stats.running.pop(fxn)

        return result

    def init_job_tracker(class_instance):
        # Initialize job_stats if it doesn't exist
        if not hasattr(class_instance, 'job_stats'):
            class_instance.job_stats = JobStats()
        return class_instance.job_stats

    return wrapper


def terminator(func):
    """
    decorator designed specifically for the SubnetScanner class,
    helps facilitate termination of a job
    """
    def wrapper(*args, **kwargs):
        scan = args[0] # aka self
        if not scan.running:
            return
        return func(*args, **kwargs)



    return wrapper
    


import webview
from .app import start_webserver_dameon
from ..libraries.runtime_args import RuntimeArgs



def start_webview(args: RuntimeArgs) -> None:
    # Start Flask server in a separate thread
    start_webserver_dameon(args)

    # Start the Pywebview window
    webview.create_window('LANscape', f'http://127.0.0.1:{args.port}')
    webview.start()


    
if __name__ == "__main__":
    # Start Flask server in a separate thread
    start_webview(True)





from ..libraries.ip_parser import get_address_count
import webview, time

import warnings

from concurrent.futures import ThreadPoolExecutor, Future
def right_size_subnet(subnet: str):
    """
    Used to improve speed of test time
    """
    if get_address_count(subnet) > 500:
        parts = subnet.split('/')
        ip = parts[0]
        mask = int(parts[1])
        mask += 1
        return right_size_subnet(f"{ip}/{mask}")
    return subnet

def webview_client(title, url):
    def decorator(func):
        def wrapper(*args, **kwargs):
            baseclass = args[0]

            window = webview.create_window(title, url)

            # Create a Future object to communicate results/exceptions
            future = Future()

            # Define the function to run in the secondary thread
            def test_function():
                # disable resource warning occuring in py >=3.13
                warnings.filterwarnings("ignore", category=ResourceWarning)

                try:
                    # Call the decorated function
                    func(baseclass, window, **kwargs)
                    future.set_result("Completed Successfully")
                except Exception as e:
                    # Set the exception in the Future
                    future.set_exception(e)
                finally:
                    # Close the WebView window
                    if window is not None:
                        window.destroy()

            # Define a function to start the thread
            def start_test_function():
                # Run the function in a secondary thread
                with ThreadPoolExecutor() as executor:
                    executor.submit(test_function)

            # Start the WebView and execute the thread alongside it
            webview.start(start_test_function)

            # Retrieve the result or exception from the Future
            future.result()

        return wrapper
    return decorator



from flask import request, jsonify
from . import api_bp
from ....libraries.net_tools import get_all_network_subnets
from ....libraries.ip_parser import parse_ip_input
from ....libraries.errors import SubnetTooLargeError
import traceback


@api_bp.route('/api/tools/subnet/test')
def test_subnet():
    subnet = request.args.get('subnet')
    if not subnet: return jsonify({'valid': False, 'msg': 'Subnet cannot be blank', 'count': -1})
    try:
        ips = parse_ip_input(subnet)
        length = len(ips)
        return jsonify({'valid': True, 'msg': f"{length} IP{'s' if length > 1 else ''}", 'count': length})
    except SubnetTooLargeError:
        return jsonify({'valid': False, 'msg': 'subnet too large', 'error': traceback.format_exc(), 'count': -1})
    except:
        return jsonify({'valid': False, 'msg': 'invalid subnet', 'error': traceback.format_exc(), 'count': -1})
    
@api_bp.route('/api/tools/subnet/list')
def list_subnet():
    """
    list all interface subets
    """
    try: 
        return jsonify(get_all_network_subnets())
    except:
        return jsonify({'error': traceback.format_exc()})